/*
 * Example 3 during tutorials on April 5, 2016.
 * Modified from example given by Prof. Manzara.
 */
public class BankAccount {
	// Account balance; this is where data corruption happens if you do not have synchronized methods.
	private double balance = 0.0;
	
	/**
	 * Deposit money to the account.
	 * @param amount The amount to deposit.
	 */
	public synchronized void deposit(double amount){
		this.balance += amount;
	}
	
	/**
	 * Withdraw money from the account.
	 * To see the race condition, run the program, then remove "synchronized" and run it again.
	 * When your method is not synchronized, keep running it and you will see the balance can
	 * either be $0 or -$100, depending on which line of code happened first (unpredictable).
	 * @param amount The amount to withdraw.
	 */
	public void withdraw(double amount){
		// We first check if there is enough money to withdraw.
		if(this.balance >= amount){
			// At this point, there is enough money to withdraw.
			System.out.println("Balance is " + this.balance + ", withdrawing $" + amount + ".");
			
			// Here we wait for a second (which is a long time for your CPU, enough to execute another withdraw method).
			// We add the one second here to increase the chances of data corruption (creating a race condition), for
			// illustration purposes.
			try{
				Thread.sleep(1000);
			}
			catch(InterruptedException e){ // using sleep() method means we need to handle InterruptedException
				System.out.println("Interrupted");
			}
			
			// After waiting, we decrease the amount. However, while waiting, another withdraw transaction may have occurred.
			// Since this transaction did not decrease the amount yet, the other withdraw transaction would have checked
			// the balance and found the $100 still there, before proceeding to withdraw the money. At this point, the balance
			// may actually be $0. Because we checked earlier, we still think there is $100 left, and we also withdraw $100.
			this.balance -= amount;
			
			// At this point, you are likely to have -$100 if your method is not synchronized.
			System.out.println("You have withdrawn $" + amount + ", your remaining balance is $" + this.balance + ".");
		}
		else{
			// If your method is synchronized, it will LOCK all the variables it uses until the method completes. That means
			// a second withdraw transaction cannot start (and check the balance) before the first one has completed the
			// withdrawal. The second transaction will then check, find that there is no money, and print out the error below.
			System.out.println("Tried to withdraw " + amount + ", insufficient funds.");
		}
	}
	
	/**
	 * Check the account balance.
	 * @return The account balance.
	 */
	public synchronized double getBalance(){
		return this.balance;
	}
}
